import functools
import os

from flask import Blueprint, Flask, send_from_directory

from ._backend import Backend


class StaticBackend(Backend):
    @staticmethod
    def init_app(flasket, rootpath, options=None) -> "StaticBackend":
        return StaticBackend(flasket, rootpath, options)

    def __init__(self, flasket, rootpath, options=None):
        super().__init__(flasket, rootpath)

        self._init_rootpath()
        self._init_blueprint()
        self._init_redirect(options)
        self._flasket._register_backend(self)

    def _init_rootpath(self) -> None:
        rootpath = self._rootpath
        if os.path.exists(os.path.join(rootpath, "static")):
            filepath = os.path.join(rootpath, "static")
            self._basepath = filepath
            self.logger.info(f'Reading static files from "{self._basepath}" directory.')
        elif os.path.exists(os.path.join(rootpath, "htdocs")):
            filepath = os.path.join(rootpath, "htdocs")
            self._basepath = filepath
            self.logger.info(f'Reading static files from "{self._basepath}" directory.')
        else:
            self._basepath = "/non-existent"
            self.logger.warning("Neither 'static' nor 'htdocs' directories in root path exist.")

    def _init_blueprint(self):
        # Use None for static_folder to prevent adding a automatic "/<filename>" route
        self._flask = Flask(__name__, root_path=self._basepath, static_folder=None, template_folder=None)
        blueprint = Blueprint(
            "static",
            __name__,
            root_path=self._basepath,
            static_folder=None,
            template_folder=None,
        )
        blueprint.add_url_rule("/<path:path>", view_func=self.send_file)
        self._flask.register_blueprint(blueprint)

    def _init_redirect(self, options):
        redirect_to = options.get("redirect_to", "/app/")
        redirect_code = options.get("redirect_code", 308)
        redirect_fn = functools.partial(self._flask.redirect, location=redirect_to, code=redirect_code)
        redirect_fn.__name__ = "redirect"
        blueprint = Blueprint(
            "redirect",
            __name__,
            root_path=self._rootpath,
            static_folder=None,
            template_folder=None,
        )
        blueprint.add_url_rule("/", view_func=redirect_fn)
        self._flask.register_blueprint(blueprint)

    @staticmethod
    def name() -> str:
        return "static"

    @property
    def prefix(self) -> str:
        return "/"

    def send_file(self, path: str) -> object:
        return send_from_directory(path=path, directory=self._basepath)



import logging
import multiprocessing
import typing as t

from flask.logging import default_handler
from gunicorn.app.base import BaseApplication
from torxtools.pathtools import expandpath

__all__: t.List[str] = [
    "GunicornApplication",
]


class GunicornApplication(BaseApplication):  # type: ignore[misc]
    # .venv/lib/python3.9/site-packages/gunicorn/app/base.py
    """
    Gunicorn WSGI application class
    """

    @staticmethod
    def _get_cpu_workers(workers: int) -> int:
        """
        Return the number of workers gunicorn should start with.
        Impose a max limit to avoid overloading the host.

        Parameters
        ----------
        workers: int

            number of workers configured

        Returns
        -------
        int
        """
        if workers > 0:
            return workers

        # Do not overload the host with a default value too large
        # if it's a container, it could report host cpus
        cpus = max((multiprocessing.cpu_count() * 2), 1)
        if workers == 0:
            return min(cpus, 8)
        return cpus

    # pylint: disable=abstract-method
    # http://docs.gunicorn.org/en/stable/custom.html
    def __init__(
        self,
        flasket,
        cfg: t.Dict[str, t.Any],
    ) -> None:
        """
        Initialize a Flasket application, and configure Gunicorn

        Parameters
        ----------
        cfg: dict

            dictionary containing configuration for the application
        """
        self._flasket = flasket
        # Rework the configuration
        cfg = cfg["server"]
        cfg["workers"] = self._get_cpu_workers(cfg.get("workers", 0))
        self.options = {
            "bind": cfg.get("listen", "localhost") + ":" + str(cfg.get("port", 8080)),
            "pidfile": expandpath(cfg.get("pidfile")),
            "workers": cfg["workers"],
            "accesslog": "/dev/null",
            "errorlog": "/dev/null",
        }

        flasket.logger.info(f"Starting {cfg['workers']} workers...")
        super().__init__()

        # Use the Flask logger to log identically between flask and gunicorn
        # We have to specify a null FileStream handler, we'll remove it on first call
        # by using the _gunicorn identifier
        logger = logging.getLogger("gunicorn.error")
        default_handler._gunicorn = False
        logger.addHandler(default_handler)

    def init(self, parser: t.Any, opts: t.Any, args: t.Any) -> None:
        """
        Overloaded abstract method.
        """
        raise NotImplementedError

    def load_config(self) -> None:
        """
        Overloaded abstract method.

        Load Gunicorn configuration ourselves.
        """
        # Pass the configuration down to gunicorn.app.base.BaseApplication
        cfg = {k: v for k, v in self.options.items() if k in self.cfg.settings and v is not None}
        for k, v in cfg.items():
            self.cfg.set(k.lower(), v)

    def load(self) -> "Flasket":
        """
        Overloaded abstract method.

        Return a callable __call__ WSGI application
        """
        return self._flasket



import os

from .clients import client
from .decorators import endpoint
from .flasket import Flasket
from .templates import template_global

__all__ = [
    "Flasket",
]


rootpath = os.path.dirname(__file__)
